from mock import patch
import pytest

import coresender
from coresender.requests.core import CoresenderClient


def _get_email_structure(data):
    return {
       'from': {
           'email': data.get('from_email', None),
           'name': data.get('from_name', None),
       },
       'to': [
            {
                'email': data.get('to_email', None),
                'name': data.get('to_name', None),
            }
       ],
       'subject': data.get('subject', None),
       'body': {
           'html': data.get('body_html', None),
           'text': data.get('body_text', None),
       },
       'custom_id': data.get('custom_id', None),
       'custom_id_unique': data.get('custom_id_unique', False),
       'track_opens': data.get('track_opens', False),
       'track_clicks': data.get('track_clicks', False),
       'list_id': data.get('list_id', None),
       'list_unsubscribe': data.get('list_unsubscribe', None),
    }


def _get_emails_structures(data):
    ret = [_get_email_structure(item) for item in data]
    return ret


@pytest.mark.asyncio
async def test_simple_email_missing_required_params(mocker):
    rq = coresender.SendEmail()
    with pytest.raises(coresender.errors.CoresenderError):
        rsp = await rq.simple_email(to_email='to@example.com', subject='test', body='test')

    with pytest.raises(coresender.errors.CoresenderError):
        rsp = await rq.simple_email(from_email='from@example.com', subject='test', body='test')


@patch('coresender.responses.SendEmail')
@pytest.mark.asyncio
async def test_simple_email_send(_, cs_ctx, mocker):
    rq = coresender.SendEmail()
    assert type(rq) is coresender.SendEmail

    client = mocker.patch.object(CoresenderClient(cs_ctx), 'send')

    rq.set_client(client)

    rsp = await rq.simple_email(from_email='from@example.com', to_email='to@example.com', subject='test', body='test')

    client.send.assert_awaited_once()


def test_add_to_batch():
    rq = coresender.SendEmail()
    assert type(rq) is coresender.SendEmail

    emails = [
        {
            'from_email': 'from@example.com',
            'to_email': 'to@example.com',
            'subject': 'test 1',
        },
        {
            'from_email': 'from@example.com',
            'to_email': 'to@example.com',
            'subject': 'test 2',
        },
    ]
    rq.add_to_batch(**emails[0])
    rq.add_to_batch(**emails[1])

    assert rq._emails == _get_emails_structures(emails)


@pytest.mark.asyncio
async def test_batch_send(cs_ctx, mocker):
    cl = mocker.patch.object(CoresenderClient(cs_ctx), 'send')
    mocker.patch('coresender.responses.SendEmail')

    rq = coresender.SendEmail()
    assert type(rq) is coresender.SendEmail

    rq.set_client(cl)

    emails = [
        {
            'from_email': 'from@example.com',
            'to_email': 'to@example.com',
            'subject': 'test 1',
        },
        {
            'from_email': 'from@example.com',
            'to_email': 'to@example.com',
            'subject': 'test 2',
        },
    ]
    rq.add_to_batch(**emails[0])
    rq.add_to_batch(**emails[1])

    rsp = await rq.execute()

    cl.send.assert_awaited_once()
    assert not rq._emails



class CoresenderError(Exception):
    pass


class CoresenderApiError(CoresenderError):
    def __init__(self, response_code, msg):
        self.response_code = response_code
        self.msg = msg

    def __str__(self):
        return self.msg


class AuthorizationError(CoresenderApiError):
    pass


class ValidationError(CoresenderApiError):
    pass


class ApiLogicError(CoresenderApiError):
    pass


class EntityNotFoundError(ApiLogicError):
    pass


class EntityExistsError(ApiLogicError):
    pass



__all__ = ['get_handler']

import logging
from typing import Callable, Optional

import httpx

from .. import errors

_logger = logging.getLogger('coresender')
_error_handlers = {}


def get_handler(rsp: httpx.Response) -> Optional[Callable[[], None]]:
    handler = None
    http_code = str(rsp.status_code)
    if http_code in _error_handlers:
        handler = _error_handlers[http_code]
    elif http_code[0] in _error_handlers:
        handler = _error_handlers[http_code[0]]
    else:
        # if it's not json or there is no `code` key - both situations handle in same way
        try:
            data = rsp.json()
            handler = _error_handlers.get(data['data']['code'], None)
        except:
            pass

    if not handler and http_code[0] != '2':
        handler = _error_handlers['default']

    if not handler:
        return

    return handler(rsp.status_code, rsp)


class ErrorHandler:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        global _error_handlers
        _error_handlers[cls.register_code] = cls

    def __init__(self, http_code: int, rsp: httpx.Response):
        self._http_code = http_code
        self._rsp = rsp


class ErrorHandlerDefault(ErrorHandler):
    register_code = 'default'

    def __call__(self):
        data = self._rsp.json()
        _logger.error("Coresender API response code is %s and data: %s", self._http_code, data)
        raise errors.CoresenderApiError(data['data']['code'], "Coresender API request failed, http status code: %s" % self._http_code)


class ErrorHandler401(ErrorHandler):
    register_code = '401'

    def __call__(self):
        _logger.error("Coresender API authorization error: [%s] %s", self._http_code, self._rsp.json())
        data = self._rsp.json()['data']['errors'][0]

        raise errors.AuthorizationError(data['code'], data['description'])


class ErrorHandler422(ErrorHandler):
    register_code = '422'

    def __call__(self):
        _logger.error("Coresender API validation error: [%s] %s", self._http_code, self._rsp.json())
        data = self._rsp.json()['data']

        msg = []
        for error in data['errors']:
            msg.append('field %s' % error['field'])
            items = ['%s: %s' % (item['code'], item['description']) for item in error['errors']]
            if items:
                msg[-1] += ': ' + ', '.join(items)

        raise errors.ValidationError(data['code'], '. '.join(msg))


class ErrorHandler409(ErrorHandler):
    register_code = '409'

    _exceptions = {
        'ENTITY_EXISTS': errors.EntityExistsError,
    }

    def __call__(self):
        _logger.error("Coresender API error: [%s] %s", self._http_code, self._rsp.json())
        data = self._rsp.json()['data']

        # there can be messing 'code' in data or missing code in _exceptions, both handled in same way
        try:
            exc_class = self._exceptions[data['code']]
        except KeyError:
            exc_class = errors.ApiLogicError

        msg = ['%s: %s' % (error['code'], error['description']) for error in data['errors']]
        raise exc_class(data['code'], '. '.join(msg))



import pytest

from coresender.context import CoresenderContext
from coresender.requests.core import CoresenderClient


@pytest.fixture
def cs_ctx():
    acc_key = 'aaa'
    acc_id = 'bbb'

    ctx = CoresenderContext()
    ctx.sending_account_id=acc_id
    ctx.sending_account_key=acc_key

    return ctx


@pytest.fixture
def cs_client(cs_ctx):
    return CoresenderClient(cs_ctx)



import coresender
from coresender import context


def test_context_is_empty():
    assert context.get_context() == None


def test_setup_context():
    acc_key = 'aaa'
    acc_id = 'bbb'
    coresender.init(sending_account_key=acc_key, sending_account_id=acc_id)

    ctx = context.get_context()
    assert type(ctx) is context.CoresenderContext
    assert ctx.sending_account_id == acc_id
    assert ctx.sending_account_key == acc_key



import pytest

from coresender import http_error_handlers as handlers
from coresender import errors


def test_default_error(mocker):
    rsp = mocker.patch('httpx.Response')
    rsp.status_code = 447
    rsp_json = {'data': {'code': '', 'errors': [{'code': 'A', 'description': 'desc'}]}}
    rsp.json = mocker.MagicMock(return_value=rsp_json)

    handler = handlers.get_handler(rsp)
    assert type(handler) is handlers.ErrorHandlerDefault

    with pytest.raises(errors.CoresenderApiError):
        handler()


def test_auth_error(mocker):
    rsp = mocker.patch('httpx.Response')
    rsp.status_code = 401
    rsp_json = {'data': {'code': '', 'errors': [{'code': 'AUTHORIZATION_ERROR', 'description': 'Invalid credentials'}]}}
    rsp.json = mocker.MagicMock(return_value=rsp_json)

    handler = handlers.get_handler(rsp)
    assert type(handler) is handlers.ErrorHandler401

    with pytest.raises(errors.AuthorizationError):
        handler()


def test_validation_error(mocker):
    rsp = mocker.patch('httpx.Response')
    rsp.status_code = 422
    rsp_json = {
        'data': {
            'code': '',
            'errors': [
                {
                    'field': 'email',
                    'errors': [
                        {
                            'code': 'INVALID_EMAIL',
                            'description': 'invalid email address',
                        }
                    ]
                }
            ]
        }
    }
    rsp.json = mocker.MagicMock(return_value=rsp_json)

    handler = handlers.get_handler(rsp)
    assert type(handler) is handlers.ErrorHandler422

    with pytest.raises(errors.ValidationError):
        handler()



__all__ = ["CoresenderContext", "get_context", "set_context"]

from typing import Optional


_ctx: Optional['CoresenderContext'] = None


class CoresenderContext:
    def __init__(self):
        self.token = None
        self.token_storage = None
        self.token_storage_params = {}
        self.token_storage_handler = None
        self.username = None
        self.password = None
        self.sending_account_key = None
        self.sending_account_id = None

    def __repr__(self):
        return ('<CoresenderContext token="%s", token_storage="%s", username="%s", password="***", '
               'sending_account_id="%s", sending_account_key="***">') % (
            self.token,
            self.token_storage,
            self.username,
            self.sending_account_id,
        )


def get_context() -> Optional[CoresenderContext]:
    return _ctx


def set_context(ctx: CoresenderContext):
    global _ctx
    _ctx = ctx



__all__ = ["init"]
__version__ = '1.0.2'

import logging
import os

from . import token
from . import context
from . import errors
from .requests import *


_logger = logging.getLogger('coresender')
_logger.addHandler(logging.NullHandler())


def init(*, sending_account_key: str = None, sending_account_id: str = None, debug: bool = False):

    ctx = context.CoresenderContext()
    ctx.sending_account_key = sending_account_key or os.environ.get('CORESENDER_SENDING_API_KEY')
    ctx.sending_account_id = sending_account_id or os.environ.get('CORESENDER_SENDING_API_ID')

    context.set_context(ctx)

    if debug or os.environ.get('CORESENDER_DEBUG', '').lower() in ('yes', 'true', '1'):
        configure_debug_logger()


def configure_debug_logger():
    import sys

    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'))
    _logger.addHandler(handler)
    _logger.setLevel(logging.DEBUG)



from .core import CoresenderApiResponse


class SendEmailResponse:
    __slots__ = ('message_id', 'custom_id', 'status', 'errors')

    def __init__(self, data: dict):
        self.message_id = data['message_id']
        self.custom_id = data['custom_id']
        self.status = data['status']
        self.errors = data['errors'] or ''

    def __repr__(self):
        r = ', '.join(['%s="%s"' % (item, getattr(self, item)) for item in self.__slots__])
        r = '<SendEmailResponse ' + r + '>'
        return r


class SendEmail(CoresenderApiResponse):
    def __init__(self, http_status, data):
        self.entries = [SendEmailResponse(item) for item in data['data']]
        self.http_status = http_status

    @property
    def all_accepted(self):
        return self.http_status == 200

    def __repr__(self):
        return '<SendEmail entries=%r>' % (self.entries, )

    def __iter__(self):
        return iter(self.entries)



__all__ = ['LoginMethod', 'CoresenderClient', 'CoresenderApiRequest']

import base64
import enum
import logging
from abc import abstractmethod
from typing import Optional, Union
from urllib.parse import quote_plus

import httpx

from .. import __version__, errors
from ..token import OAuth2Token
from ..context import CoresenderContext, get_context
from ..http_error_handlers import get_handler as get_error_handler


_logger = logging.getLogger('coresender')
_client: Optional['CoresenderClient'] = None


class LoginMethod(enum.Enum):
    oauth2 = 'oauth2'
    api_key = 'api_key'


class CoresenderClientAuth(httpx.Auth):
    def __init__(self,
        oauth2_token_required: bool, api_key_required: bool,
        access_token: str,
        sending_account_key: str, sending_account_id: str
    ):
        self.oauth2_token_required = oauth2_token_required
        self.api_key_required = api_key_required
        self.access_token = access_token
        self.sending_account_key = sending_account_key
        self.sending_account_id = sending_account_id

    def auth_flow(self, request):
        if self.oauth2_token_required:
            request.headers['Authorization'] = 'Bearer ' + self.access_token
            _logger.debug("Auth: using oauth2 token")
        elif self.api_key_required:
            token = base64.b64encode(b'%s:%s' % (self.sending_account_id.encode(), self.sending_account_key.encode()))
            request.headers['Authorization'] = 'Basic ' + token.decode()
            _logger.debug("Auth: using basic auth")

        yield request


class CoresenderClient:
    def __init__(self, ctx: CoresenderContext):
        self._ctx = ctx

    @classmethod
    def _url_encode(cls, params: dict) -> str:
        r = []
        for key, value in params.items():
            key = quote_plus(key)
            if isinstance(value, (tuple, list)):
                for single_value in value:
                    if single_value is not None:
                        r.append('%s[]=%s' % (key, quote_plus(str(single_value).encode())))
            elif value is not None:
                r.append('%s=%s' % (key, '' if value is None else quote_plus(str(value).encode())))

        r = '&'.join(r)
        return r

    @classmethod
    def _build_url(cls, url: str, params: dict) -> str:
        if params:
            params = cls._url_encode(params)
            if params:
                sign = '&' if '?' in url else '?'
                url += sign + params
        return url

    async def login(self, force: bool = False) -> None:
        if self._ctx.token and self._ctx.token.is_valid() and not force:
            _logger.debug("Reusing saved OAuth2 token")
            return

        login = Login()
        data = {
            "grant_type": "password",
            "email": self._ctx.username,
            "password": self._ctx.password,
        }

        rsp = await self.send(login.api_method, login.get_full_url(), data)
        json_response = rsp.json()
        if 'access_token' in json_response:
            self._ctx.token = OAuth2Token.from_rq_json(json_response)
            _logger.info("Logged in as %s", data['email'])

            self._ctx.token_storage_handler.save(self._ctx.token)
        else:
            raise errors.CoresenderError("Unrecognized response from Coresender API: [%s] %s" % (rsp.status_code, json_response))

    async def send(self, method: str, url: str, data: dict = None, options: dict = None) -> httpx.Response:
        if not options:
            options = {}

        headers = {
            'User-Agent': 'coresender-sdk-python/%s' % __version__,
            'Accept': 'application/json',
        }
        auth = CoresenderClientAuth(
            oauth2_token_required=options.get('oauth2_token_required', False),
            api_key_required=options.get('api_key_required', False),
            access_token=(self._ctx.token.access_token if self._ctx.token else None),
            sending_account_key=self._ctx.sending_account_key,
            sending_account_id=self._ctx.sending_account_id,
        )

        if auth.oauth2_token_required and not auth.access_token:
            await self.login()
            auth.access_token = self._ctx.token.access_token

        if 'headers' in options:
            headers.update(options['headers'])

        url = self._build_url(url, options.get('query_params', {}))
        _logger.debug("Sending %s to %s with headers %s and data %s", method, url, headers, data)

        async with httpx.AsyncClient() as cl:
            rsp = await cl.request(method, url, headers=headers, json=data, auth=auth)

        _logger.debug("Coresender API response is [%s] %s", rsp.status_code, rsp.text)

        error_handler = get_error_handler(rsp)
        if error_handler:
            error_handler()

        return rsp


class CoresenderApiRequest:
    _api_version: str = None
    _api_proto: str = 'https'
    _api_host: str = 'api.coresender.com'
    _api_port: int = 443
    _api_method: str = None
    _api_method_uri: str = None
    _login_required: bool = None
    _login_method: LoginMethod = None

    _client: CoresenderClient = None

    @classmethod
    def client(cls) -> Optional[CoresenderClient]:
        if not cls._client:
            ctx = get_context()
            cls._client = CoresenderClient(ctx)

        return cls._client

    @classmethod
    def set_client(cls, client: CoresenderClient) -> None:
        cls._client = client

    async def send(self, *, data: Union[list, dict] = None, qs: dict = None, headers: dict = None):
        query_params = self.get_query_params() or {}
        if qs:
            query_params.update(qs)

        options = {
            'query_params': query_params,
            'headers': headers or {},
            'api_key_required': (self.login_required and self.login_method is LoginMethod.api_key),
            'oauth2_token_required': (self.login_required and self.login_method is LoginMethod.oauth2),
        }

        query_data = data or self.to_json()

        api_rsp = await self.client().send(self._api_method, self.get_full_url(), query_data, options)

        return api_rsp

    @abstractmethod
    async def execute(self) -> 'CoresenderApiRequest':
        raise NotImplementedError()

    def get_full_url(self) -> str:
        url_prefix = '%(proto)s://%(host)s:%(port)d/v%(version)s/%(uri)s' % {
            'proto': self._api_proto,
            'host': self._api_host,
            'port': self._api_port,
            'version': self._api_version,
            'uri': self._api_method_uri,
        }
        return url_prefix

    def to_json(self) -> Optional[Union[dict, list]]:
        if not hasattr(self, '_to_json'):
            return None
        return self._to_json()

    def get_query_params(self) -> dict:
        return {}

    @property
    def api_method(self) -> str:
        return self._api_method

    @property
    def login_required(self) -> bool:
        return self._login_required

    @property
    def login_method(self) -> Optional[LoginMethod]:
        return self._login_method


class Login(CoresenderApiRequest):
    _api_method: str = 'POST'
    _api_version: str = '1'
    _api_method_uri: str = 'login'
    _login_required: LoginMethod = False



#!/usr/bin/env python

import os.path
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

requires = [
    'httpx==0.12.0',
]
test_requirements = [
    'mock',
    'pytest-mock',
    'pytest-asyncio',
    'pytest>=3',
]

with open('README.md', 'r', encoding='utf-8') as fh:
    readme = fh.read()

setup(
    name='coresender',
    version='1.0.2',
    description='Coresender API for sending email',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Marcin Sztolcman',
    author_email='opensource@coresender.com',
    url='https://coresender.com',
    packages=find_packages(),
    package_data={'': ['LICENSE', ]},
    package_dir={'coresender': 'coresender'},
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=requires,
    license='MIT',
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Communications :: Email',
        'Topic :: Communications :: Email :: Mail Transport Agents',
        'Environment :: Web Environment',
    ],
    tests_require=test_requirements,
    extras_require={
    },
    project_urls={
        'API Documentation': 'https://coresender.com/docs/api',
        'Source': 'https://github.com/coresender/coresender-sdk-py',
    },
    keywords="coresender api email smtp mta transactional",
)



__all__ = ["OAuth2Token", "get_storage_handler"]

import datetime
import json
import logging
import pathlib
from abc import abstractmethod
from typing import Optional, Type

from . import errors


_logger = logging.getLogger('coresender')
_storage_handlers = {}


def get_storage_handler(name: str) -> Type['OAuth2TokenStorage']:
    try:
        token_storage = _storage_handlers[name]
    except KeyError:
        raise errors.CoresenderError("Unknown token storage: %s" % name)

    return token_storage


class OAuth2Token:
    __slots__ = ('_token_type', '_expires_on', '_access_token', '_refresh_token')

    def __init__(self):
        self._token_type = None
        self._expires_on = None
        self._access_token = None
        self._refresh_token = None

    def is_valid(self) -> bool:
        return self.expires_on > datetime.datetime.now()

    @property
    def token_type(self) -> Optional[str]:
        return self._token_type

    @token_type.setter
    def token_type(self, value: str):
        self._token_type = str(value)

    @property
    def expires_on(self) -> Optional[datetime.datetime]:
        return self._expires_on

    @expires_on.setter
    def expires_on(self, value: datetime.datetime):
        self._expires_on = value

    def expire_in(self, value: str):
        self.expires_on = datetime.datetime.now() + datetime.timedelta(seconds=int(value))

    @property
    def access_token(self) -> Optional[str]:
        return self._access_token

    @access_token.setter
    def access_token(self, value: str):
        self._access_token = value

    @property
    def refresh_token(self) -> Optional[str]:
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, value: str):
        self._refresh_token = value

    @classmethod
    def from_rq_json(cls, data: dict) -> 'OAuth2Token':
        r = cls()
        r.access_token = data['access_token']
        r.refresh_token = data['refresh_token']
        r.token_type = data['token_type']
        r.expire_in(data['expires_in'])
        return r

    @classmethod
    def from_json(cls, data: dict) -> 'OAuth2Token':
        r = cls()
        r.access_token = data['access_token']
        r.refresh_token = data['refresh_token']
        r.token_type = data['token_type']
        r.expires_on = datetime.datetime.fromisoformat(data['expires_on'])
        return r

    def to_json(self) -> dict:
        json = {
            'access_token': self.access_token,
            'refresh_token': self.refresh_token,
            'token_type': self.token_type,
            'expires_on': self.expires_on.isoformat(),
        }
        return json

    def __repr__(self):
        r = ', '.join(['%s="%s"' % (item[1:], getattr(self, item)) for item in self.__slots__])
        r = '<OAuth2Token ' + r + '>'
        return r


class OAuth2TokenStorage:
    def __init_subclass__(cls, storage_name: str, **kwargs):
        super().__init_subclass__(**kwargs)

        global _storage_handlers
        _storage_handlers[storage_name] = cls

    def __init__(self, params: dict):
        self.params = params

    @abstractmethod
    def read(self) -> Optional[OAuth2Token]:
        raise NotImplementedError()

    @abstractmethod
    def save(self, token: OAuth2Token) -> None:
        raise NotImplementedError()


class OAuth2TokenFileStorage(OAuth2TokenStorage, storage_name='file'):
    default_storage_path = '~/.coresender.token'

    def _get_path(self):
        path = self.params.get('path', self.default_storage_path)
        path = pathlib.Path(path).expanduser()
        return path

    def read(self) -> Optional[OAuth2Token]:
        path = self._get_path()
        if not path.exists():
            return

        with path.open('r') as fh:
            try:
                token = json.load(fh)
            except Exception as exc:
                _logger.exception("Cannot read saved token from %s", path, exc_info=exc)
                return

        token = OAuth2Token.from_json(token)
        _logger.debug("Success reading cached token data from %s", path)
        return token

    def save(self, token: OAuth2Token) -> None:
        path = self._get_path()
        with path.open('w') as fh:
            json.dump(token.to_json(), fh)
            _logger.debug("OAuth2Token successfully dumped into %s" % path)



__all__ = ["BodyType", "SendEmail"]

import enum
from typing import List, Dict

from .core import CoresenderApiRequest, LoginMethod
from .. import responses
from .. import errors


class BodyType(enum.Enum):
    text = 'text'
    html = 'html'


class SendEmail(CoresenderApiRequest):
    _api_version: str = '1'
    _api_method: str = 'POST'
    _api_method_uri: str = 'send_email'
    _login_required: bool = True
    _login_method: LoginMethod = LoginMethod.api_key

    def __init__(self):
        self._emails = []

    def _validate_email(self, email: dict):
        if not email['from']['email']:
            raise errors.CoresenderError('No sender address specified')
        if not email['to'][0]['email']:
            raise errors.CoresenderError('No recipient address specified')

    def add_to_batch(self,
       from_email: str, from_name: str = None,
       to: List[Dict[str, str]] = None,
       to_email: str = None, to_name: str = None,
       subject: str = None, body_html: str = None, body_text: str = None,
       custom_id: str = None, custom_id_unique: bool = False,
       track_opens: bool = False, track_click: bool = False,
       list_id: str = None, list_unsubscribe: str = None
    ) -> None:
        if not to:
            to = [{'email': to_email, 'name': to_name}]

        email = {
            "from": {
                "email": from_email,
                "name": from_name,
            },
            "to": to,
            "subject": subject,
            "body": {
                "html": body_html,
                "text": body_text,
            },
            "custom_id": custom_id,
            "custom_id_unique": custom_id_unique,
            "track_opens": track_opens,
            "track_clicks": track_click,
            "list_id": list_id,
            "list_unsubscribe": list_unsubscribe,
        }

        self._validate_email(email)

        self._emails.append(email)

    async def execute(self) -> responses.SendEmail:
        if not self._emails:
            raise errors.CoresenderError("No emails scheduled to send")

        api_rsp = await self.send()

        rsp = responses.SendEmail(api_rsp.status_code, api_rsp.json())

        self._emails.clear()

        return rsp

    async def simple_email(self,
        from_email: str = None, to_email: str = None,
        subject: str = None,
        body: str = None, *, body_type: BodyType = BodyType.text
    ) -> responses.SendEmailResponse:
        email = {
            "from": {
                "email": from_email,
            },
            "to": [{'email': to_email, }],
            "subject": subject,
            "body": {
                body_type.value: body,
            },
        }
        self._validate_email(email)

        api_rsp = await self.send(data=[email])

        rsp = responses.SendEmail(api_rsp.status_code, api_rsp.json())

        return rsp.entries[0]

    def _to_json(self):
        return self._emails
